 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * Memory access functions
  *
  * Copyright 1996 Bernd Schmidt
  */

static __inline__ ULONG do_get_mem_long(ULONG *a)
{
   ULONG retval;

   __asm__ ("bswapl %0" : "=r" (retval) : "0" (*a) : "cc");

   return retval;
}

static __inline__ ULONG do_get_mem_word(UWORD *a)
{
   ULONG retval;

   __asm__ ("xorl %k0,%k0\n\tmovw %w1,%w0\n\trolw $8,%w0" : "=&r" (retval) : "m" (*a) : "cc");

   return retval;
}

#define do_get_mem_byte(a) ((ULONG)*(a))

static __inline__ void do_put_mem_long(ULONG *a, ULONG v)
{
   __asm__ ("bswapl %0" : "=r" (v) : "0" (v) : "cc");
   *a = v;
}

static __inline__ void do_put_mem_word(UWORD *a, UWORD v)
{
   __asm__ ("rolw $8,%0" : "=r" (v) : "0" (v) : "cc");
   *a = v;
}

#define do_put_mem_byte(a,v) (*(a) = (v))

#if 0
static __inline__ ULONG call_mem_get_func(mem_get_func func, CPTR addr)
{
   ULONG result;

   __asm__("call *%1" : "=a" (result) : "r" (func), "a" (addr) : "cc", "edx", "ecx");

   return result;
}

static __inline__ void call_mem_put_func(mem_put_func func, CPTR addr, ULONG v)
{
   __asm__("call *%2" : : "a" (addr), "d" (v), "r" (func) : "cc", "eax", "edx", "ecx", "memory");
}

#else

   #define call_mem_get_func(func,addr) ((*func)(addr))
   #define call_mem_put_func(func,addr,v) ((*func)(addr,v))

#endif

#undef NO_INLINE_MEMORY_ACCESS
#undef MD_HAVE_MEM_1_FUNCS

#if defined(MEMFUNCS_DIRECT_REQUESTED) && USER_PROGRAMS_BEHAVE > 0
   #define MD_HAVE_MEM_1_FUNCS

   #if USER_PROGRAMS_BEHAVE > 1
      #define MAYBE_MASK_ADDR_24BIT(a) (a)
   #else
      #define MAYBE_MASK_ADDR_24BIT(a) ((a) & 0xFFFFFF)
   #endif

   #define longget_1(addr)    (do_get_mem_long ((ULONG *)(address_space + MAYBE_MASK_ADDR_24BIT(addr))))
   #define wordget_1(addr)    (do_get_mem_word ((UWORD *)(address_space + MAYBE_MASK_ADDR_24BIT(addr))))
   #define byteget_1(addr)    (do_get_mem_byte (address_space + MAYBE_MASK_ADDR_24BIT(addr)))
   #define longput_1(addr, l) (do_put_mem_long ((ULONG*)(address_space + MAYBE_MASK_ADDR_24BIT(addr)), (l)))
   #define wordput_1(addr, w) (do_put_mem_word ((UWORD *)(address_space + MAYBE_MASK_ADDR_24BIT(addr)), (w)))
   #define byteput_1(addr, b) (do_put_mem_byte (address_space + MAYBE_MASK_ADDR_24BIT(addr), (b)))
 
   #define DIRECT_MEMFUNCS_SUCCESSFUL
#else /* ! MEMFUNCS_DIRECT_REQUESTED */
   #ifdef MD_HAVE_MEM_1_FUNCS

static __inline__ ULONG longget_1(CPTR addr)
{
   ULONG result;

   __asm__ ("andl   $0x00FFFFFF,%1    \n\t"
            "cmpb   $0,(%1,%3)        \n\t"
            "leal   1f,%%ecx          \n\t"
            "je     longget_stub      \n\t"
            "addl   address_space,%1  \n\t"
            "movl   (%1),%0           \n\t"
            "bswapl %0                \n\t"
            "1:" : "=c" (result), "=d" (addr) : "1" (addr), "r" (good_address_map) : "cc");

   return result;
}

static __inline__ ULONG wordget_1(CPTR addr)
{
   ULONG result;

   __asm__ ("andl   $0x00FFFFFF,%1     \n\t"
            "cmpb   $0,(%1,%3)         \n\t"
            "leal   1f,%%ecx           \n\t"
            "je     wordget_stub       \n\t"
            "addl   address_space,%1   \n\t"
            "movzwl (%1),%0            \n\t"
            "rolw $8,%w0               \n\t"
            "1:" : "=c" (result), "=d" (addr) : "1" (addr), "r" (good_address_map) : "cc");

   return result;
}

static __inline__ ULONG byteget_1(CPTR addr)
{
   ULONG result;

   __asm__ ("andl   $0x00FFFFFF,%1    \n\t"
            "cmpb   $0,(%1,%3)        \n\t"
            "leal   1f,%%ecx          \n\t"
            "je     byteget_stub      \n\t"
            "addl   address_space,%1  \n\t"
            "movzbl (%1),%0           \n\t"
            "1:" : "=c" (result), "=d" (addr) : "1" (addr), "r" (good_address_map) : "cc");

   return result;
}

static __inline__ void longput_1(CPTR addr, ULONG l)
{
   __asm__ __volatile__("andl   $0x00FFFFFF,%0    \n\t"
                        "cmpb   $0,(%0,%3)        \n\t"
                        "leal   1f,%%ecx          \n\t"
                        "je     longput_stub      \n\t"
                        "addl   address_space,%0  \n\t"
                        "bswapl %1                \n\t"
                        "movl   %1,(%0)           \n\t"
                        "1:" : "=d" (addr), "=b" (l) : "0" (addr), "r" (good_address_map), "1" (l) : "cc", "memory", "ecx");
}

static __inline__ void wordput_1(CPTR addr, ULONG w)
{
   __asm__ __volatile__("andl $0x00FFFFFF,%0    \n\t"
                        "cmpb $0,(%0,%3)        \n\t"
                        "leal 1f,%%ecx          \n\t"
                        "je   wordput_stub      \n\t"
                        "addl address_space,%0  \n\t"
                        "rolw $8,%1             \n\t"
                        "movw %w1,(%0)          \n\t"
                        "\t1:" : "=d" (addr), "=b" (w) : "0" (addr), "r" (good_address_map), "1" (w) : "cc", "memory", "ecx");
}

static __inline__ void byteput_1(CPTR addr, ULONG b)
{
   __asm__ __volatile__("andl $0x00FFFFFF,%0     \n\t"
                        "tcmpb $0,(%0,%3)        \n\t"
                        "tleal 1f,%%ecx          \n\t"
                        "tje byteput_stub        \n\t"
                        "taddl address_space,%0  \n\t"
                        "tmovb %b1,(%0)          \n\t"
                        "t1:" : "=d" (addr), "=b" (b) : "0" (addr), "r" (good_address_map), "1" (b) : "cc", "memory", "ecx");
}

#endif

#endif /* ! MEMFUNCS_DIRECT_REQUESTED */
