/*
 *  IBM sound device driver
 *
 *  (c) 1998, Yoav Shadmi
 *
 */

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"

#include "sound.h"
#include "ram.h"
#include "via.h"
#include "memory2.h"

#include "IBM-SND.h"

#include <stdio.h>
#include <memory.h>
#include <go32.h>
#include <dpmi.h>

#include "allegro.h"

// Sound driver
extern Sound_Ty Sound;
static int stream=0;
int Installed;

UBYTE Mono_8_SoundBuffer[370];
UBYTE Left_8_SoundBuffer[185];          // Left speaker stereo buffer
UBYTE Right_8_SoundBuffer[185];         // Right speaker stereo buffer

SAMPLE *mono_8;                         // 8bit mono
SAMPLE *left_8;                         // 8bit stereo left
SAMPLE *right_8;                        // 8bit stereo right

// Detect and install the sound driver
int Sound_Install (void)
{
   if (install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, NULL) != 0) {
      allegro_exit();
      fprintf(stderr, "vMac: Can't initialize sound driver.\n");
      return 1;
   } else
      fprintf(stderr, "vMac: Sound driver initialized.\n");

   #ifdef STEREO
      left_8  = malloc(sizeof(SAMPLE));
      right_8 = malloc(sizeof(SAMPLE));

      left_8->bits       = 8;
      left_8->freq       = 11025;
      left_8->priority   = 255;
      left_8->len        = 185;
      left_8->loop_start = 0;
      left_8->loop_end   = 185;
      left_8->param      = -1;
      left_8->data       = (void *)Left_8_SoundBuffer;

      right_8->bits       = 8;
      right_8->freq       = 11025;
      right_8->priority   = 255;
      right_8->len        = 185;
      right_8->loop_start = 0;
      right_8->loop_end   = 185;
      right_8->param      = -1;
      right_8->data       = (void *)Right_8_SoundBuffer;
   #else
      mono_8  = malloc(sizeof(SAMPLE));

      mono_8->bits       = 8;
      mono_8->freq       = 22050;
      mono_8->priority   = 255;
      mono_8->len        = 370;
      mono_8->loop_start = 0;
      mono_8->loop_end   = 370;
      mono_8->param      = -1;
      mono_8->data       = (void *)Mono_8_SoundBuffer;
   #endif

   Installed = 1;
   return 0;
}

// Uninstall sound driver
void Sound_Uninstall (void)
{
   #ifdef STEREO
      destroy_sample(left_8);
      destroy_sample(right_8);
   #else
      destroy_sample(mono_8);
   #endif

   remove_sound();
   Installed = 0;
   return;
}

// set sound blaster master volume
void Sound_SetVolume (UBYTE vol)
{
   set_volume((int)vol<<5, -1);
}

void Sound_Pause (void)
{
   stop_sample(mono_8);
   stream=0;
}

// Read in data.  Play it if the VIA sound bit is set.
#ifdef STEREO
void Sound_Play(void)
{
   ULONG kSound_Address;
   int i;

   if (Sound.Buffer==0)
      kSound_Address = kSnd_Alt_Buffer;
   else 
      kSound_Address = kSnd_Main_Buffer;

   Sound_SetVolume(Sound.Volume);

   if (Sound.Enable) {
      for (i=0; i<185; i++)
          Left_8_SoundBuffer[i] = Right_8_SoundBuffer[i] = 128;

      if (stream==1) {                          // stop stream
         stop_sample(left_8);
         stop_sample(right_8);
         stream=0;
      }
   } else {
      int x=0;
      for (i=0; i<740; i+=4)  {
         Left_8_SoundBuffer[x]  = get_byte(kSound_Address+i);
         Right_8_SoundBuffer[x] = get_byte(kSound_Address+i+2);
         x++;
      }

      if (stream==0) {                          // start stream
         play_sample(left_8, Sound.Volume << 5, 0  , 1000, -1);
         play_sample(right_8, Sound.Volume << 5, 255, 1000, -1);
         stream=1;
      }
   }
}
#else
void Sound_Play(void)
{
   ULONG kSound_Address;
   int i;

   if (Sound.Buffer==0)
      kSound_Address = kSnd_Alt_Buffer;
   else 
      kSound_Address = kSnd_Main_Buffer;

   Sound_SetVolume(Sound.Volume);

   if (Sound.Enable) {
      for (i=0; i<370; i++)
          Mono_8_SoundBuffer[i] = 128;

      if (stream==1) {                          // stop stream
         stop_sample(mono_8);
         stream=0;
      }
   } else {
      for (i=0; i<370; i++)
         Mono_8_SoundBuffer[i] = get_byte(kSound_Address+(i<<1));

      if (stream==0) {             // start stream
         play_sample(mono_8, Sound.Volume << 5, 128, 1000, -1);
         stream=1;
      }
   }
}
#endif

void Sound_LockDriver (void)
{
   _go32_dpmi_lock_code (Sound_Play, (unsigned int)Sound_LockDriver - (unsigned int)Sound_Play);

   _go32_dpmi_lock_data (&Installed, (unsigned int) sizeof(Installed));
   _go32_dpmi_lock_data (&stream,    (unsigned int) sizeof(stream));
   _go32_dpmi_lock_data (mono_8,     (unsigned int) sizeof(SAMPLE));
   _go32_dpmi_lock_data (left_8,     (unsigned int) sizeof(SAMPLE));
   _go32_dpmi_lock_data (right_8,    (unsigned int) sizeof(SAMPLE));

   return;
}
