/*
 *  IBM mouse/keyboard device driver
 *
 *  (c) 1998, Yoav Shadmi
 *
 */

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "prefs.h"

#include "Memory.h"
#include "memory2.h"
#include "custom2.h"
#include "readcpu.h"
#include "newcpu.h"

#include "screen.h"
#include "keyboard.h"
#include "sony.h"

#include "keys.h"
#include "mackbd.h"

#include "IBM-UI.h"
#include "IBM-VGA.h"

#include <stdio.h>
#include <dos.h>
#include <go32.h>
#include <dpmi.h>
#include <signal.h>
#include <conio.h>

#include "allegro.h"

// Mouse device driver
int MouseInstalled = 0;
extern volatile int buttonstate;
volatile int newmousecounters = 0;
volatile int lastmx = 0, lastmy = 0;
extern volatile void Mouse_Update2 (void);
extern volatile void MouseEnd      (void);

static void Mouse_Handler (int flags)
{
   int x,y;

   if ((flags & MOUSE_FLAG_MOVE) || (flags & MOUSE_FLAG_LEFT_DOWN) || (flags & MOUSE_FLAG_LEFT_UP)) {
      get_mouse_mickeys(&x, &y);

      lastmx=lastmx+x;
      lastmy=lastmy+y;

      if (thePrefs.Big==2) {
         if (lastmx>Right2-1) { lastmx=Right2-1 ; position_mouse(lastmx, lastmy);}
         if (lastmx<Left)     { lastmx=Left     ; position_mouse(lastmx, lastmy);}
         if (lastmy>Bottom2-1){ lastmy=Bottom2-1; position_mouse(lastmx, lastmy);}
         if (lastmy<Top)      { lastmy=Top      ; position_mouse(lastmx, lastmy);}
      } else {
         if (lastmx>511)      { lastmx=511 ; position_mouse(lastmx, lastmy); }
         if (lastmx<0)        { lastmx=0   ; position_mouse(lastmx, lastmy); }
         if (lastmy>341)      { lastmy=341 ; position_mouse(lastmx, lastmy); }
         if (lastmy<0)        { lastmy=0   ; position_mouse(lastmx, lastmy); }
      }

      buttonstate=(mouse_b & 1);

      Mouse_Update2();
   }
}

int  Mouse_Install (void)
{
   __dpmi_regs IntRegs;

   if (MouseInstalled)
      return(1);

   IntRegs.x.ax = 0;
   __dpmi_int(0x33, &IntRegs);

   if (IntRegs.x.ax==0) {
      VGA_EnterDebug();
      _set_screen_lines(25);
      _quit_=1;
      fprintf(stderr, "vMac: Mouse driver not present.\n");
      DumpMessages();
      raise(SIGUSR1);
      return(0);
   }

   install_mouse();

   mouse_callback=(void *)Mouse_Handler;

   buttonstate = 0;
   lastmx = lastmy = 0;

   MouseInstalled = 1;
   return(1);
}

void Mouse_Uninstall (void)
{
   if (!MouseInstalled)
      return;

   mouse_callback=NULL;
   
   remove_mouse();
   MouseInstalled = 0;
}

void Mouse_LockDriver (void)
{
   _go32_dpmi_lock_code(Mouse_Update2, (unsigned int)MouseEnd      - (unsigned int)Mouse_Update2);
   _go32_dpmi_lock_code(Mouse_Handler, (unsigned int)Mouse_Install - (unsigned int)Mouse_Handler);

   return;
}


// Keyboard device driver
int KeyBoardInstalled = 0;
volatile int Leds;
int KeyState[256];
_go32_dpmi_seginfo OldKeyboardHandler, NewKeyboardHandler;
volatile int C_Press=0, N_Press=0, S_Press=0;
unsigned char EscapeKeys[128] = {
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   96,  97,  0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   98,  0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   99,
    100, 0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   101, 102,
    103, 104, 0,   105, 0,   106, 0,   107,
    108, 109, 110, 111, 0,   0,   0,   0,
    0,   0,   0,   125, 126, 127, 0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
};

static void Keyboard_Handler(void)
{
   extern int use_debugger;

   static int IsEscape = 0, CapsLockCanChange = 1;
   int ScanCode, NewState, MacKey, Mods=0;
   char MacScanCode=0xFF;

   ScanCode = inportb(0x60);

   if (ScanCode) {
      if (ScanCode == 0xe0) {
         IsEscape = 1;
         outportb(0x20, 0x20);
         return;
      }

      // PAUSE Key
      if (IsEscape > 1) {
         IsEscape++;

         if (IsEscape == 7)
            IsEscape = 0;

         outportb(0x20, 0x20);
         return;
      }

      if (ScanCode == 0xe1) {
         IsEscape = 2;
         outportb(0x20, 0x20);
         return;
      }
   
      NewState = !(ScanCode & 0x80);
      ScanCode = ScanCode & 0x7f;
  
      if (IsEscape) {
         ScanCode = EscapeKeys[ScanCode];
         IsEscape = 0;
      }

      switch (ScanCode) {
         case _NUMLOCK:
            if (N_Press==0) {
               if (Leds & 0x02) {
                  Leds &= ~0x02;
               } else {
                  Leds |= 0x02;
               }
 
               do { } while ((inportb(0x64) & 0x02)!=0);
               outportb(0x60, 0xed);
               do { } while ((inportb(0x64) & 0x02)!=0);
               outportb(0x60, Leds);

               N_Press=1;
            } else {
               N_Press=0;
            }
            break;

         case _SCRLLCK:
            if (S_Press==0) {
               if (Leds & 0x01) {
                  Leds &= ~0x01;
               } else {
                  Leds |= 0x01;
               }
 
               do { } while ((inportb(0x64) & 0x02)!=0);
               outportb(0x60, 0xed);
               do { } while ((inportb(0x64) & 0x02)!=0);
               outportb(0x60, Leds);

               S_Press=1;
            } else {
               S_Press=0;
            }
            break;

         case _CAPSLCK:
            if (C_Press==0) {
               if (Leds & 0x04) {
                  Leds &= ~0x04;
               } else {
                  Leds |= 0x04;
               }
 
               do { } while ((inportb(0x64) & 0x02)!=0);
               outportb(0x60, 0xed);
               do { } while ((inportb(0x64) & 0x02)!=0);
               outportb(0x60, Leds);

               C_Press=1;
            } else {
               C_Press=0;
            }
            break;

         case _F1:
            if (!Sony_Disk_Inserted(1)) {
               Insert_Disk=1;
               GUI=1;
               regs.spcflags |= SPCFLAG_BRK;
               use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
               quit_program = 1;
            }
            break;
      
         case _F2:
            if (!Sony_Disk_Inserted(2)) {
               Insert_Disk=2;
               GUI=0;
               regs.spcflags |= SPCFLAG_BRK;
               use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
               quit_program = 1;
            }
            break;

         case _F3:
            if (!Sony_Disk_Inserted(3)) {
               Insert_Disk=3;
               GUI=0;
               regs.spcflags |= SPCFLAG_BRK;
               use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
               quit_program = 1;
            }
            break;

         case _F4:
            if (!Sony_Disk_Inserted(4)) {
               Insert_Disk=4;
               GUI=0;
               regs.spcflags |= SPCFLAG_BRK;
               use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
               quit_program = 1;
            }
            break;

         case _F8: {
            extern int Dump;

            GUI=1;
            Dump=1;
            regs.spcflags |= SPCFLAG_BRK;
            use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
            quit_program = 1;
            break;
         }

         case _F9:
            GUI=0;
            regs.spcflags |= SPCFLAG_BRK;
            use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
            quit_program = 1;
            break;

         case _F10:
            GUI=0;
            regs.spcflags |= SPCFLAG_BRK;
            use_debugger = 0;       /* 1 - go to debug ; 0 - exit */
            quit_program = 1;
            break;

         case _F11:
            INTERRUPT=1;
            GUI=1;
            regs.spcflags |= SPCFLAG_BRK;
            use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
            quit_program = 1;
            break;

         case _F12:
            RESET_SW=1;
            GUI=1;
            regs.spcflags |= SPCFLAG_BRK;
            use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
            quit_program = 1;
            break;
      }

      outportb(0x20, 0x20);

      MacKey = IBM2Mac_reg[ScanCode];

      if (MacKey == 0xFFFF)
         return;

      if (MacKey == SC_CLK) {
         if (NewState) {
            if (CapsLockCanChange) {
               NewState = !KeyState[ScanCode];
               CapsLockCanChange = 0;
            } else
               return;
         } else {
            CapsLockCanChange = 1;
            return;
         }
      }

      if (NewState && KeyState[ScanCode])
         NewState = 2;
      KeyState[ScanCode] = NewState;


      if (MacKey < 0xFF00) {
         Mods=0;

         if (KeyState[_LALT] || KeyState[_RALT]) {
            Mods |= 0x0100;
            Keyboard_UpdateKeyMap(kCommandKeyMap, 1); 
         }

         if (KeyState[58]) {                      // Caps-Lock
            Mods |= 0x0400;
            Keyboard_UpdateKeyMap(kCapsLockKeyMap, 1); 
            if (((MacKey & 0x00FF)>=97) && ((MacKey & 0x00FF)<=122))
               MacKey-=32;                        // Capital Letters
         }

         if (KeyState[42] || KeyState[54]) {      // Shift
            Mods |= 0x0200;
            Keyboard_UpdateKeyMap(kShiftKeyMap, 1); 
            if (((MacKey & 0x00FF)>=97) && ((MacKey & 0x00FF)<=122)) {
               MacKey-=32;
            } else {
               switch (MacKey & 0x00FF) {
            /*`*/ case 0x60: MacKey = (MacKey&0xFF00)+0x7E; break;
            /*1*/ case 0x31: MacKey = (MacKey&0xFF00)+0x21; break;
            /*2*/ case 0x32: MacKey = (MacKey&0xFF00)+0x40; break;
            /*3*/ case 0x33: MacKey = (MacKey&0xFF00)+0x23; break;
            /*4*/ case 0x34: MacKey = (MacKey&0xFF00)+0x24; break;           
            /*5*/ case 0x35: MacKey = (MacKey&0xFF00)+0x25; break;
            /*6*/ case 0x36: MacKey = (MacKey&0xFF00)+0x5E; break;
            /*7*/ case 0x37: MacKey = (MacKey&0xFF00)+0x26; break;
            /*8*/ case 0x38: MacKey = (MacKey&0xFF00)+0x2A; break;
            /*9*/ case 0x39: MacKey = (MacKey&0xFF00)+0x28; break;
            /*0*/ case 0x30: MacKey = (MacKey&0xFF00)+0x29; break;
            /*-*/ case 0x2D: MacKey = (MacKey&0xFF00)+0x5F; break;
            /*=*/ case 0x3D: MacKey = (MacKey&0xFF00)+0x2B; break;
            /*\*/ case 0x5C: MacKey = (MacKey&0xFF00)+0x7C; break;
            /*[*/ case 0x5B: MacKey = (MacKey&0xFF00)+0x7B; break;
            /*]*/ case 0x5D: MacKey = (MacKey&0xFF00)+0x7D; break;
            /*;*/ case 0x3B: MacKey = (MacKey&0xFF00)+0x3A; break;
            /*'*/ case 0x27: MacKey = (MacKey&0xFF00)+0x22; break;
            /*,*/ case 0x2C: MacKey = (MacKey&0xFF00)+0x3C; break;
            /*.*/ case 0x2E: MacKey = (MacKey&0xFF00)+0x3E; break;
            /*?*/ case 0x2F: MacKey = (MacKey&0xFF00)+0x3F; break;
               }
            }
         }

         if (KeyState[29] || KeyState[97]) {
            Mods |= 0x0800;
            Keyboard_UpdateKeyMap(kOptionKeyMap, 1); 
         }

         switch (NewState) {
            case 0:
               Keyboard_UpdateKeyMap(MacKey >> 8, 0);
               PC_PostEvent (4, MacKey, Mods);
               break;
            case 1:
               Keyboard_UpdateKeyMap(MacKey >> 8, 1);
               PC_PostEvent (3, MacKey, Mods);
               break;
            case 2:
               PC_PostEvent (5, MacKey, Mods);
               break;
         }
      }
   }
}

int Keyboard_Install(void)
{
   int i;
   char b;

   if (KeyBoardInstalled)
      return(1);

   _go32_dpmi_get_protected_mode_interrupt_vector(9, &OldKeyboardHandler);

   NewKeyboardHandler.pm_offset = (int)Keyboard_Handler;
   if (_go32_dpmi_allocate_iret_wrapper(&NewKeyboardHandler) != 0) {
      VGA_EnterDebug();
      _set_screen_lines(25);
      _quit_=1;
      fprintf(stderr, "vMac: Can't allocate enough memory for keyboard handler.\n");
      DumpMessages();
      raise(SIGUSR1);
      return(0);
   }

   if (_go32_dpmi_set_protected_mode_interrupt_vector(9, &NewKeyboardHandler) != 0) {
      VGA_EnterDebug();
      _set_screen_lines(25);
      _quit_=1;
      fprintf(stderr, "vMac: Can't install keyboard handler.\n");
      DumpMessages();
      raise(SIGUSR1);
      return(0);
   }

   for (i = 0; i < 256; i++)
      KeyState[i] = 0;

   Leds=0;

   KeyBoardInstalled = 1;

   return(1);
}

void Keyboard_Uninstall(void)
{
   if (!KeyBoardInstalled)
      return;

   _go32_dpmi_set_protected_mode_interrupt_vector(9, &OldKeyboardHandler);
   _go32_dpmi_free_iret_wrapper(&NewKeyboardHandler);

   do { } while ((inportb(0x64) & 0x02)!=0);
   outportb(0x60, 0xed);
   do { } while ((inportb(0x64) & 0x02)!=0);
   outportb(0x60, 0);

   KeyBoardInstalled = 0;
}

void Keyboard_LockDriver (void)
{
   _go32_dpmi_lock_code(Keyboard_Handler, (unsigned int)Keyboard_Install-(unsigned int)Keyboard_Handler);
   _go32_dpmi_lock_data(KeyState, sizeof(KeyState));
   _go32_dpmi_lock_data(&OldKeyboardHandler, sizeof(OldKeyboardHandler));
   _go32_dpmi_lock_data(&NewKeyboardHandler, sizeof(NewKeyboardHandler));

   return;
}
