/*
 *  IBM VGA device driver
 *
 *  (c) 1998, Yoav Shadmi
 *
 */

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "prefs.h"

#include "memory2.h"
#include "Memory.h"

#include "RAM.h"
#include "screen.h"

#include "IBM-TIME.h"
#include "IBM-UI.h"
#include "IBM-VGA.h"

#include <go32.h>
#include <dpmi.h>
#include <conio.h>


// VGA device driver
UWORD private_VRAM_Main [screenBytes2];
UWORD *VRAM_Main = private_VRAM_Main;

UBYTE vPage2 = 1;
int OldVGAMode;
int InGraphics = 0;
int UseScreen = 0;

int VGA_VRAMInit (void)
{
   char *MacVideo, *VGAVideo;
   int i;
   int rowBytes;
   int Bottom;

   if (thePrefs.Big) {
      for (i=0; i<screenBytes2; i++)
         VRAM_Main[i] = 0xFFFF;              // clear to black

      MacVideo = (char *) VRAM_Main;

      if (thePrefs.Big==1) {
         VGAVideo = (char *) RegScreen+0x1598;
         rowBytes = rowBytes1;
         Bottom   = Bottom1;
      } else {
         VGAVideo = (char *) RegScreen;
         rowBytes = rowBytes2;
         Bottom   = Bottom2;
      }

      for (i=0; i<Bottom; i++, MacVideo+=rowBytes, VGAVideo+=80)
         asm __volatile__ (
            "      movw %0, %%fs                   \t\n"
            "_NextPixel:                           \t\n"
            "      movl (%%esi), %%ebx             \t\n"
            "      rorl $0x08, %%ebx               \t\n"
            "      movl %%ebx, %%eax               \t\n"
            "      andl $0x00ff00ff, %%eax         \t\n"
            "      roll $0x10, %%ebx               \t\n"
            "      andl $0xff00ff00, %%ebx         \t\n"
            "      orl  %%ebx, %%eax               \t\n"
            "      notl %%eax                      \t\n"
            "      movl %%eax, %%fs:(%%edi)        \t\n"
            "      addl $0x4, %%esi                \t\n"
            "      addl $0x4, %%edi                \t\n"
            "      loop _NextPixel                 \t\n"
            :
            : "g" (_dos_ds), "S" (MacVideo), "D" (VGAVideo), "c" (rowBytes/4)
            : "%eax", "%ebx", "%esi", "%edi", "%ecx", "cc", "memory"
        );
   }
   return 0;
}

inline void VGA_ScreenDraw (void)
{
/*
   char    *VGAVideo, *MacVideo;
   int     i;

   if (vPage2 == 1)
      MacVideo = (char *) RAM_Xlate (kOld_Main_Buffer);
   else
      MacVideo = (char *) RAM_Xlate (kOld_Alternate_Buffer);

   VGAVideo = (char *) RegScreen+0x1598;

   for (i=0; i<342; i++, MacVideo+=64, VGAVideo+=80)
      asm __volatile__ (
         "      movw %0, %%fs                   \t\n"
         ".NextPixel:                           \t\n"
         "      movl (%%esi), %%eax             \t\n"
         "      notl %%eax                      \t\n"
         "      movl %%eax, %%fs:(%%edi)        \t\n"
         "      addl $0x4, %%esi                \t\n"
         "      addl $0x4, %%edi                \t\n"
         "      loop .NextPixel                 \t\n"
         :
         : "g" (_dos_ds), "S" (MacVideo), "D" (VGAVideo), "c" (16)
         : "%eax", "%esi", "%edi", "%ecx", "cc", "memory"
      );
*/
   return;
}

void VGA_ViewScreen (void)
{
   __dpmi_regs Regs;
   int rowBytes;
   int Bottom;

   Regs.h.ah = 0x00;
   Regs.h.al = 0x11;                            // 0x11
   __dpmi_int(0x10, &Regs);
   InGraphics = 1;
   UseScreen = 0;

   if (thePrefs.Big) {
      char *MacVideo, *VGAVideo;
      int  i;

      MacVideo = (char *) VRAM_Main;

      if (thePrefs.Big==1) {
         VGAVideo = (char *) RegScreen+0x1598;
         rowBytes = rowBytes1;
         Bottom   = Bottom1;
      } else {
         VGAVideo = (char *) RegScreen;
         rowBytes = rowBytes2;
         Bottom   = Bottom2;
      }

      for (i=0; i<Bottom; i++, MacVideo+=rowBytes, VGAVideo+=80)
         asm __volatile__ (
            "      movw %0, %%fs                   \t\n"
            "..NextPixel:                          \t\n"
            "      movl (%%esi), %%ebx             \t\n"
            "      rorl $0x08, %%ebx               \t\n"
            "      movl %%ebx, %%eax               \t\n"
            "      andl $0x00ff00ff, %%eax         \t\n"
            "      roll $0x10, %%ebx               \t\n"
            "      andl $0xff00ff00, %%ebx         \t\n"
            "      orl  %%ebx, %%eax               \t\n"
            "      notl %%eax                      \t\n"
            "      movl %%eax, %%fs:(%%edi)        \t\n"
            "      addl $0x4, %%esi                \t\n"
            "      addl $0x4, %%edi                \t\n"
            "      loop ..NextPixel                \t\n"
            :
            : "g" (_dos_ds), "S" (MacVideo), "D" (VGAVideo), "c" (rowBytes/4)
            : "%eax", "%ebx", "%esi", "%edi", "%ecx", "cc", "memory"
         );
   } else {
      char *MacVideo, *VGAVideo;
      int  i;

      MacVideo = (char *) RAM_Xlate (kOld_Main_Buffer);
      VGAVideo = (char *) RegScreen+0x1598;

      for (i=0; i<342; i++, MacVideo+=64, VGAVideo+=80)
         asm __volatile__ (
            "      movw %0, %%fs                   \t\n"
            ".NextPixel:                           \t\n"
            "      movl (%%esi), %%eax             \t\n"
            "      notl %%eax                      \t\n"
            "      movl %%eax, %%fs:(%%edi)        \t\n"
            "      addl $0x4, %%esi                \t\n"
            "      addl $0x4, %%edi                \t\n"
            "      loop .NextPixel                 \t\n"
            :
            : "g" (_dos_ds), "S" (MacVideo), "D" (VGAVideo), "c" (16)
            : "%eax", "%esi", "%edi", "%ecx", "cc", "memory"
         );
   }

   getch();
   VGA_EnterDebug();
   return;
}

int VGA_InstallGraphics (void)
{
   __dpmi_regs Regs;
   int rowBytes;
   int Bottom;

   Regs.h.ah = 0x00;
   Regs.h.al = 0x11;
   __dpmi_int(0x10, &Regs);

   InGraphics = 1;
   UseScreen = 0;

   Keyboard_Install();

   if (thePrefs.Big) {
      char *MacVideo, *VGAVideo;
      int  i;

      MacVideo = (char *) VRAM_Main;

      if (thePrefs.Big==1) {
         VGAVideo = (char *) RegScreen+0x1598;
         rowBytes = rowBytes1;
         Bottom   = Bottom1;
      } else {
         VGAVideo = (char *) RegScreen;
         rowBytes = rowBytes2;
         Bottom   = Bottom2;
      }

      for (i=0; i<Bottom; i++, MacVideo+=rowBytes, VGAVideo+=80)
         asm __volatile__ (
            "      movw %0, %%fs                   \t\n"
            "NextPixel:                            \t\n"
            "      movl (%%esi), %%ebx             \t\n"
            "      rorl $0x08, %%ebx               \t\n"
            "      movl %%ebx, %%eax               \t\n"
            "      andl $0x00ff00ff, %%eax         \t\n"
            "      roll $0x10, %%ebx               \t\n"
            "      andl $0xff00ff00, %%ebx         \t\n"
            "      orl  %%ebx, %%eax               \t\n"
            "      notl %%eax                      \t\n"
            "      movl %%eax, %%fs:(%%edi)        \t\n"
            "      addl $0x4, %%esi                \t\n"
            "      addl $0x4, %%edi                \t\n"
            "      loop NextPixel                  \t\n"
            :
            : "g" (_dos_ds), "S" (MacVideo), "D" (VGAVideo), "c" (rowBytes/4)
            : "%eax", "%ebx", "%esi", "%edi", "%ecx", "cc", "memory"
         );
   } else {
      char *MacVideo, *VGAVideo;
      int  i;

      MacVideo = (char *) RAM_Xlate (kOld_Main_Buffer);
      VGAVideo = (char *) RegScreen+0x1598;

      for (i=0; i<342; i++, MacVideo+=64, VGAVideo+=80)
         asm __volatile__ (
            "      movw %0, %%fs                   \t\n"
            "__NextPixel:                          \t\n"
            "      movl (%%esi), %%eax             \t\n"
            "      notl %%eax                      \t\n"
            "      movl %%eax, %%fs:(%%edi)        \t\n"
            "      addl $0x4, %%esi                \t\n"
            "      addl $0x4, %%edi                \t\n"
            "      loop __NextPixel                \t\n"
            :
            : "g" (_dos_ds), "S" (MacVideo), "D" (VGAVideo), "c" (16)
            : "%eax", "%esi", "%edi", "%ecx", "cc", "memory"
         );
   }

   return 1;
}

void VGA_EnterDebug(void)
{
   __dpmi_regs Regs;

   if (InGraphics == 0)
      return;
   InGraphics = 0;

   Regs.x.ax = 0x1C;
   Regs.h.ah = 0x00;
   Regs.h.al = 0x03;
   __dpmi_int(0x10, &Regs);

   Keyboard_Uninstall();

   _set_screen_lines(50);
   clrscr();
   DumpMessages();
}

void VGA_UninstallGraphics(void)
{
   __dpmi_regs Regs;

   // Uninstall screen
   Regs.x.ax = 0x1C;
   Regs.h.ah = 0x00;
   Regs.h.al = OldVGAMode;
   __dpmi_int(0x10, &Regs);

   DumpMessages();
}

void VGA_LockDriver (void)
{
   _go32_dpmi_lock_code(VGA_VRAMInit,   (unsigned int)VGA_ScreenDraw - (unsigned int)VGA_VRAMInit);
   _go32_dpmi_lock_code(VGA_ScreenDraw, (unsigned int)VGA_ViewScreen - (unsigned int)VGA_ScreenDraw);

   _go32_dpmi_lock_data(VRAM_Main,      (unsigned int)sizeof(private_VRAM_Main));

   return;
}

UBYTE VRAM_Main_GetB (CPTR addr)
{
   return VRAM_Main [addr >> 1] >> (addr & 1 ? 0 : 8);
}

UWORD VRAM_Main_GetW (CPTR addr)
{
   return VRAM_Main [addr >> 1];
}

ULONG VRAM_Main_GetL (CPTR addr)
{
   return ((ULONG) VRAM_Main [addr >> 1] << 16) | VRAM_Main [(addr >> 1)+1];
}

void VRAM_Main_PutB (CPTR addr, UBYTE in)
{
   if (!(addr & 1)) {
      VRAM_Main [addr >> 1] = (VRAM_Main [addr >> 1] & 0xff) | (((UWORD) in) << 8);
   } else {
      VRAM_Main [addr >> 1] = (VRAM_Main [addr >> 1] & 0xff00) | in;
   }

   if (thePrefs.Big==1)  addr+=0x1598;

   asm __volatile__ (
      "movw     %0, %%fs                \t\n"
      "notb     %%al                    \t\n"
      "movb     %%al, %%fs:(%%edi)      \t\n"
      :
      : "g" (_dos_ds), "a" (in), "D" (RegScreen+addr)
      : "%eax", "%edi", "cc", "memory"
   );

   return;
}

void VRAM_Main_PutW (CPTR addr, UWORD in)
{
   VRAM_Main [addr >> 1] = in;

   if (thePrefs.Big==1)  addr+=0x1598;

   asm __volatile__ (
      "movw     %0, %%fs                \t\n"
      "xchg     %%al, %%ah              \t\n"
      "notw     %%ax                    \t\n"
      "movw     %%ax, %%fs:(%%edi)      \t\n"
       :
       : "g" (_dos_ds), "a" (in), "D" (RegScreen+addr)
       : "%eax", "%edi", "cc", "memory"
   );

   return;
}

void VRAM_Main_PutL (CPTR addr, ULONG in)
{
   VRAM_Main [addr >> 1] = in >> 16;
   VRAM_Main [(addr >> 1) + 1] = (UWORD) in;

   if (thePrefs.Big==1)  addr+=0x1598;

   asm __volatile__ (
      "movw     %0, %%fs                \t\n"
      "bswap    %%eax                   \t\n"
      "notl     %%eax                   \t\n"
      "movl     %%eax, %%fs:(%%edi)     \t\n"
      :
      : "g" (_dos_ds), "a" (in), "D" (RegScreen+addr)
      : "%eax", "%edi", "cc", "memory"
   );

   return;
}

UBYTE VRAM_Alt_GetB (CPTR addr)
{
   return 0;
}

UWORD VRAM_Alt_GetW (CPTR addr)
{
   return 0;
}

ULONG VRAM_Alt_GetL (CPTR addr)
{
   return 0;
}

void VRAM_Alt_PutB (CPTR addr, UBYTE in)
{
   return;
}

void VRAM_Alt_PutW (CPTR addr, UWORD in)
{
   return;
}

void VRAM_Alt_PutL (CPTR addr, ULONG in)
{
   return;
}
