 /* 
  * Trap.c
  * Fatal exceptions handler [including m68k critical exceptions]
  *
  * vMac for MS-DOS systems
  * 
  * (c) Yoav Shadmi, 1998
  * 
  */

#include "sysconfi.h"
#include "sysdeps.h"

#include <conio.h>
#include <signal.h>
#include <stdarg.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <dpmi.h>

#include "config.h"
#include "options.h"
#include "Trap.h"
#include "Prefs.h"
#include "memory2.h"
#include "custom.h"
#include "r_newcpu.h"

#include "screen.h"

#include "IBM-UI.h"
#include "IBM-VGA.h"
#include "IBM-TIME.h"

extern char ConfigFile[];
extern int no_msg;
extern int _quit_;


void fatal_error (char *_sig)
{
   extern void allegro_exit (void);

   if (!_quit_) {
      VGA_EnterDebug();
      _set_screen_lines(25);
   }

   fprintf(stderr, "vMac: Fatal Exception: %s\n", _sig);

   Mouse_Uninstall();
   Timer_Uninstall();
   UninstallExceptionHandler();
   allegro_exit();

   PrefsSave(ConfigFile);

   no_msg=1;
   exit(1);
}


static int caught_fatal_sig = 0;

static RETSIGTYPE ExceptionHandler (int sig_num)
{
   char *sig="\0";

   if (caught_fatal_sig)
      kill (getpid (), sig_num);

   caught_fatal_sig = 1;

   switch (sig_num) {
      case SIGHUP : sig = "SIGHUP  [modem hangup].\nvMac cannot continue and will be terminated."; break;    
      case SIGQUIT: sig = "SIGQUIT [quit() requested].\nvMac cannot continue and will be terminated."; break;
      case SIGABRT: sig = "SIGABRT [abort() requested].\nvMac cannot continue and will be terminated."; break;
      case SIGSEGV: sig = "SIGSEGV [general protection fault].\nvMac cannot continue and will be terminated."; break;
      case SIGPIPE: sig = "SIGPIPE [Broken pipe].\nRestart your machine."; break;
      case SIGTERM: sig = "SIGTERM [termination requested].\nvMac cannot continue and will be terminated."; break;
      case SIGNOFP: sig = "Sorry, vMac requires an FPU co-processor.\nvMac cannot continue and will be terminated."; break;
      case SIGFPE : sig = "SIGFPE  [FPU fatal exception].\nvMac cannot continue and will be terminated."; break;
      case SIGKILL: sig = "SIGKILL. vMac cannot continue and will be terminated."; break;
      case SIGUSR1: sig = "Sorry, vMac cannot continue and will be terminated."; break;
      case SIGUSR2: sig = "The emulated m68k has encountered an illegal instruction.vMac cannot continue and will be terminated."; break;

      default:      sig = "Unknown exception.\nvMac cannot continue and will be terminated."; break;
   }

   fatal_error (sig);
}

void InstallExceptionHandler(void)
{
   signal(SIGHUP,  ExceptionHandler);
   signal(SIGQUIT, ExceptionHandler);
   signal(SIGABRT, ExceptionHandler);
   signal(SIGSEGV, ExceptionHandler);
   signal(SIGPIPE, ExceptionHandler);
   signal(SIGTERM, ExceptionHandler);
   signal(SIGNOFP, ExceptionHandler);
   signal(SIGFPE,  ExceptionHandler);
   signal(SIGKILL, ExceptionHandler);
   signal(SIGUSR1, ExceptionHandler);
   signal(SIGUSR2, ExceptionHandler);

   _go32_dpmi_lock_code(&ExceptionHandler, (unsigned long)(InstallExceptionHandler-(unsigned long)&ExceptionHandler));
   _go32_dpmi_lock_data(&caught_fatal_sig, sizeof(caught_fatal_sig));

   fprintf(stderr, "vMac: ExceptionHandler() - engaged.\n");
}

void UninstallExceptionHandler(void)
{
   signal(SIGHUP,  SIG_DFL);
   signal(SIGQUIT, SIG_DFL);
   signal(SIGABRT, SIG_DFL);
   signal(SIGSEGV, SIG_DFL);
   signal(SIGPIPE, SIG_DFL);
   signal(SIGTERM, SIG_DFL);
   signal(SIGNOFP, SIG_DFL);
   signal(SIGFPE,  SIG_DFL);
   signal(SIGKILL, SIG_DFL);
   signal(SIGUSR1, SIG_DFL);
   signal(SIGUSR2, SIG_DFL);
}
